"""Concatenate MRI time series data.

Usage:
    tsconcat bids_dir output_dir group [options]

Arguments:
    bids_dir (str): Input BIDS folder path.
    output_dir (str): Output BIDS folder path.
    analysis_level (str): Processing stage, must be "group".
"""
